# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Interface utility functions

"""

__docformat__ = 'restructuredtext'

import json
import logging
import sys
from os import listdir
from os.path import isdir
from os.path import join as opj
from os.path import (
    relpath,
    sep,
)
from time import time
from typing import TypeVar

import datalad.support.ansi_colors as ac
from datalad import cfg as dlcfg
from datalad.dochelpers import single_or_plural
from datalad.support.exceptions import CapturedException
from datalad.support.gitrepo import GitRepo
from datalad.ui import ui
# avoid import from API to not get into circular imports
from datalad.utils import (
    ensure_unicode,
    getargspec,
    path_is_subpath,
    path_startswith,
)
from datalad.utils import \
    with_pathsep as \
    _with_sep  # TODO: RF whenever merge conflict is not upon us

anInterface = TypeVar('anInterface', bound='Interface')

lgr = logging.getLogger('datalad.interface.utils')


# TODO remove
# only `drop` and `uninstall` are still using this
def handle_dirty_dataset(ds, mode, msg=None):
    """Detect and treat unsaved changes as instructed by `mode`

    Parameters
    ----------
    ds : Dataset or None
      Dataset to be inspected. Does nothing if `None`.
    mode : {'fail', 'ignore', 'save-before'}
      How to act upon discovering unsaved changes.
    msg : str or None
      Custom message to use for a potential commit.

    Returns
    -------
    None
    """
    if ds is None:
        # nothing to be handled
        return
    if msg is None:
        msg = '[DATALAD] auto-saved changes'

    # make sure that all pending changes (batched annex operations, etc.)
    # are actually reflected in Git
    if ds.repo:
        ds.repo.precommit()

    if mode == 'ignore':
        return
    elif mode == 'fail':
        if not ds.repo or ds.repo.dirty:
            raise RuntimeError('dataset {} has unsaved changes'.format(ds))
    elif mode == 'save-before':
        if not ds.is_installed():
            raise RuntimeError('dataset {} is not yet installed'.format(ds))
        from datalad.core.local.save import Save
        Save.__call__(dataset=ds, message=msg, updated=True)
    else:
        raise ValueError("unknown if-dirty mode '{}'".format(mode))


def get_tree_roots(paths):
    """Return common root paths for a set of paths

    This function determines the smallest set of common root
    paths and sorts all given paths under the respective
    root.

    Returns
    -------
    dict
      paths by root
    """
    paths_ws = [_with_sep(p) for p in paths]
    # sort all paths under their potential roots
    roots = {}
    # start from the top to get all paths down the line
    # and collate them into as few roots as possible
    for s in sorted(paths_ws):
        if any([s.startswith(r) for r in roots]):
            # this path is already covered by a known root
            continue
        # find all sub paths
        subs = [p for p in paths if p.startswith(s)]
        roots[s.rstrip(sep)] = subs
    return roots


# TODO(OPT)? YOH: from a cursory review seems like possibly an expensive function
# whenever many paths were provided (e.g. via shell glob).
# Might be worth testing on some usecase and py-spy'ing if notable portion
# of time is spent.
def discover_dataset_trace_to_targets(basepath, targetpaths, current_trace,
                                      spec, includeds=None):
    """Discover the edges and nodes in a dataset tree to given target paths

    Parameters
    ----------
    basepath : path
      Path to a start or top-level dataset. Really has to be a path to a
      dataset!
    targetpaths : list(path)
      Any non-zero number of paths that are termination points for the
      search algorithm. Can be paths to datasets, directories, or files
      (and any combination thereof).
    current_trace : list
      For a top-level call this should probably always be `[]`
    spec : dict
      `content_by_ds`-style dictionary that will receive information about the
      discovered datasets. Specifically, for each discovered dataset there
      will be an item with its path under the key (path) of the respective
      superdataset.
    includeds : sequence, optional
      Any paths given are treated as existing subdatasets, regardless of
      whether they can be found in the filesystem. Such subdatasets will appear
      under the key of the closest existing dataset in the `spec`.

    Returns
    -------
    None
      Function calls itself recursively and populates `spec` dict in-place.
      Keys are dataset paths, values are sets of subdataset paths
    """
    # convert to set for faster lookup
    includeds = includeds if isinstance(includeds, set) else \
        set() if includeds is None else set(includeds)
    # this beast walks the directory tree from a given `basepath` until
    # it discovers any of the given `targetpaths`
    # if it finds one, it commits any accumulated trace of visited
    # datasets on this edge to the spec
    valid_repo = GitRepo.is_valid_repo(basepath)
    if valid_repo:
        # we are passing into a new dataset, extend the dataset trace
        current_trace = current_trace + [basepath]
    # this edge is not done, we need to try to reach any downstream
    # dataset
    undiscovered_ds = set(t for t in targetpaths)  # if t != basepath)
    # whether anything in this directory matched a targetpath
    filematch = False
    if isdir(basepath):
        for p in listdir(basepath):
            p = ensure_unicode(opj(basepath, p))
            if not isdir(p):
                if p in targetpaths:
                    filematch = True
                # we cannot have anything below this one
                continue
            # OPT listdir might be large and we could have only few items
            # in `targetpaths` -- so traverse only those in spec which have
            # leading dir basepath
            # filter targets matching this downward path
            downward_targets = set(
                t for t in targetpaths if path_startswith(t, p))
            if not downward_targets:
                continue
            # remove the matching ones from the "todo" list
            undiscovered_ds.difference_update(downward_targets)
            # go one deeper
            discover_dataset_trace_to_targets(
                p, downward_targets, current_trace, spec,
                includeds=includeds if not includeds else includeds.intersection(
                    downward_targets))
    undiscovered_ds = [t for t in undiscovered_ds
                       if includeds and
                       path_is_subpath(t, current_trace[-1]) and
                       t in includeds]
    if filematch or basepath in targetpaths or undiscovered_ds:
        for i, p in enumerate(current_trace[:-1]):
            # TODO RF prepare proper annotated path dicts
            subds = spec.get(p, set())
            subds.add(current_trace[i + 1])
            spec[p] = subds
        if undiscovered_ds:
            spec[current_trace[-1]] = spec.get(current_trace[-1], set()).union(
                undiscovered_ds)


def get_result_filter(fx):
    """Wrap a filter into a helper to be able to accept additional
    arguments, if the filter doesn't support it already"""
    _fx = fx
    if fx and not getargspec(fx).keywords:
        def _fx(res, **kwargs):
            return fx(res)
    return _fx


def eval_results(wrapped):
    import warnings

    from datalad.interface.base import eval_results as eval_results_moved
    warnings.warn("datalad.interface.utils.eval_results is obsolete. "
                  "Use datalad.interface.base.eval_results instead",
                  DeprecationWarning)
    return eval_results_moved(wrapped)


def generic_result_renderer(res):
    if res.get('status', None) != 'notneeded':
        path = res.get('path', None)
        if path and res.get('refds'):
            try:
                path = relpath(path, res['refds'])
            except ValueError:
                # can happen, e.g., on windows with paths from different
                # drives. just go with the original path in this case
                pass
        ui.message('{action}({status}):{path}{type}{msg}{err}'.format(
            action=ac.color_word(
                res.get('action', '<action-unspecified>'),
                ac.BOLD),
            status=ac.color_status(res.get('status', '<status-unspecified>')),
            path=' {}'.format(path) if path else '',
            type=' ({})'.format(
                ac.color_word(res['type'], ac.MAGENTA)
            ) if 'type' in res else '',
            msg=' [{}]'.format(
                res['message'][0] % res['message'][1:]
                if isinstance(res['message'], tuple) else res[
                    'message'])
            if res.get('message', None) else '',
            err=ac.color_word(' [{}]'.format(
                res['error_message'][0] % res['error_message'][1:]
                if isinstance(res['error_message'], tuple) else res[
                    'error_message']), ac.RED)
            if res.get('error_message', None) and res.get('status', None) != 'ok' else ''))


# keep for legacy compatibility
default_result_renderer = generic_result_renderer


def render_action_summary(action_summary):
    ui.message("action summary:\n  {}".format(
        '\n  '.join('{} ({})'.format(
            act,
            ', '.join('{}: {}'.format(status, action_summary[act][status])
                      for status in sorted(action_summary[act])))
                    for act in sorted(action_summary))))


def _display_suppressed_message(nsimilar, ndisplayed, last_ts, final=False):
    # +1 because there was the original result + nsimilar displayed.
    n_suppressed = nsimilar - ndisplayed + 1
    if n_suppressed > 0:
        ts = time()
        # rate-limit update of suppression message, with a large number
        # of fast-paced results updating for each one can result in more
        # CPU load than the actual processing
        # arbitrarily go for a 2Hz update frequency -- it "feels" good
        if last_ts is None or final or (ts - last_ts > 0.5):
            ui.message('  [{} similar {} been suppressed; disable with datalad.ui.suppress-similar-results=off]'
                       .format(n_suppressed,
                               single_or_plural("message has",
                                                "messages have",
                                                n_suppressed, False)),
                       cr="\n" if final else "\r")
            return ts
    return last_ts


def _process_results(
        results,
        cmd_class,
        on_failure,
        action_summary,
        incomplete_results,
        result_renderer,
        result_log_level,
        allkwargs):
    # private helper pf @eval_results
    # loop over results generated from some source and handle each
    # of them according to the requested behavior (logging, rendering, ...)

    # used to track repeated messages in the generic renderer
    last_result = None
    # the timestamp of the last renderer result
    last_result_ts = None
    # counter for detected repetitions
    last_result_reps = 0
    # how many repetitions to show, before suppression kicks in
    render_n_repetitions = \
        dlcfg.obtain('datalad.ui.suppress-similar-results-threshold') \
            if sys.stdout.isatty() \
               and dlcfg.obtain('datalad.ui.suppress-similar-results') \
            else float("inf")

    for res in results:
        if not res or 'action' not in res:
            # XXX Yarik has to no clue on how to track the origin of the
            # record to figure out WTF, so he just skips it
            # but MIH thinks leaving a trace of that would be good
            lgr.debug('Drop result record without "action": %s', res)
            continue

        actsum = action_summary.get(res['action'], {})
        if res['status']:
            actsum[res['status']] = actsum.get(res['status'], 0) + 1
            action_summary[res['action']] = actsum
        ## log message, if there is one and a logger was given
        msg = res.get('message', None)
        # remove logger instance from results, as it is no longer useful
        # after logging was done, it isn't serializable, and generally
        # pollutes the output
        res_lgr = res.pop('logger', None)
        if msg and res_lgr:
            if isinstance(res_lgr, logging.Logger):
                # didn't get a particular log function, go with default
                res_lgr = getattr(
                    res_lgr,
                    default_logchannels[res['status']]
                    if result_log_level == 'match-status'
                    else result_log_level)
            msg = res['message']
            msgargs = None
            if isinstance(msg, tuple):
                msgargs = msg[1:]
                msg = msg[0]
            if 'path' in res:
                # result path could be a path instance
                path = str(res['path'])
                if msgargs:
                    # we will pass the msg for %-polation, so % should be doubled
                    path = path.replace('%', '%%')
                msg = '{} [{}({})]'.format(
                    msg, res['action'], path)
            if msgargs:
                # support string expansion of logging to avoid runtime cost
                try:
                    res_lgr(msg, *msgargs)
                except TypeError as exc:
                    raise TypeError(
                        "Failed to render %r with %r from %r: %s"
                        % (msg, msgargs, res, str(exc))
                    ) from exc
            else:
                res_lgr(msg)

        ## output rendering
        if result_renderer is None or result_renderer == 'disabled':
            pass
        elif result_renderer == 'generic':
            last_result_reps, last_result, last_result_ts = \
                _render_result_generic(
                    res, render_n_repetitions,
                    last_result_reps, last_result, last_result_ts)
        elif result_renderer in ('json', 'json_pp'):
            _render_result_json(res, result_renderer.endswith('_pp'))
        elif result_renderer == 'tailored':
            cmd_class.custom_result_renderer(res, **allkwargs)
        elif hasattr(result_renderer, '__call__'):
            _render_result_customcall(res, result_renderer, allkwargs)
        else:
            raise ValueError(f'unknown result renderer "{result_renderer}"')

        ## error handling
        # looks for error status, and report at the end via
        # an exception
        if on_failure in ('continue', 'stop') \
                and res['status'] in ('impossible', 'error'):
            incomplete_results.append(res)
            if on_failure == 'stop':
                # first fail -> that's it
                # raise will happen after the loop
                break
        yield res
    # make sure to report on any issues that we had suppressed
    _display_suppressed_message(
        last_result_reps, render_n_repetitions, last_result_ts, final=True)


def _render_result_generic(
        res, render_n_repetitions,
        # status vars
        last_result_reps, last_result, last_result_ts):
    # which result dict keys to inspect for changes to discover repetitions
    # of similar messages
    repetition_keys = set(('action', 'status', 'type', 'refds'))

    trimmed_result = {k: v for k, v in res.items() if k in repetition_keys}
    if res.get('status', None) != 'notneeded' \
            and trimmed_result == last_result:
        # this is a similar report, suppress if too many, but count it
        last_result_reps += 1
        if last_result_reps < render_n_repetitions:
            generic_result_renderer(res)
        else:
            last_result_ts = _display_suppressed_message(
                last_result_reps, render_n_repetitions, last_result_ts)
    else:
        # this one is new, first report on any prev. suppressed results
        # by number, and then render this fresh one
        last_result_ts = _display_suppressed_message(
            last_result_reps, render_n_repetitions, last_result_ts,
            final=True)
        generic_result_renderer(res)
        last_result_reps = 0
    return last_result_reps, trimmed_result, last_result_ts


def _render_result_json(res, prettyprint):
    ui.message(json.dumps(
        {k: v for k, v in res.items()
         if k not in ('logger')},
        sort_keys=True,
        indent=2 if prettyprint else None,
        default=str))


def _render_result_customcall(res, result_renderer, allkwargs):
    try:
        result_renderer(res, **allkwargs)
    except Exception as e:
        lgr.warning('Result rendering failed for: %s [%s]',
                    res, CapturedException(e))


def keep_result(res, rfilter, **kwargs):
    if not rfilter:
        return True
    try:
        if not rfilter(res, **kwargs):
            # give the slightest indication which filter was employed
            raise ValueError(
                'excluded by filter {} with arguments {}'.format(rfilter, kwargs))
    except ValueError as e:
        # make sure to report the excluded result to massively improve
        # debugging experience
        lgr.debug('Not reporting result (%s): %s', CapturedException(e), res)
        return False
    return True


def xfm_result(res, xfm):
    if not xfm:
        return res

    return xfm(res)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Provide access to stuff (html, data files) via HTTP and HTTPS

"""
import io
import re
from logging import getLogger
from time import sleep

import requests
import requests.auth
from requests.utils import parse_dict_header

from .. import __version__
from ..dochelpers import borrowkwargs
from ..log import LoggerHelper
from ..support.cookies import cookies_db
from ..support.exceptions import (
    AccessDeniedError,
    AccessFailedError,
    CapturedException,
    DownloadError,
    UnhandledRedirectError,
)
from ..support.network import (
    get_response_disposition_filename,
    get_url_filename,
    rfc2822_to_epoch,
)
from ..support.status import FileStatus
from ..ui import ui
from ..utils import (
    auto_repr,
    ensure_bytes,
    ensure_dict_from_str,
    ensure_list_from_str,
)
from .base import (
    Authenticator,
    BaseDownloader,
    DownloaderSession,
)

# at some point was trying to be too specific about which exceptions to
# catch for a retry of a download.
# from urllib3.exceptions import MaxRetryError, NewConnectionError





lgr = getLogger('datalad.http')

# Following https://meta.wikimedia.org/wiki/User-Agent_policy to provide
# extended and informative User-Agent string
DEFAULT_USER_AGENT = \
    f'DataLad/{__version__} ' \
    '(https://datalad.org; team@datalad.org) ' \
    f'python-requests/{requests.__version__}'

try:
    import requests_ftp
    _FTP_SUPPORT = True
    requests_ftp.monkeypatch_session()
except ImportError as e:
    ce = CapturedException(e)
    lgr.debug("Failed to import requests_ftp, thus no ftp support: %s", ce)
    _FTP_SUPPORT = False

if lgr.getEffectiveLevel() <= 1:
    # Let's also enable requests etc debugging

    # These two lines enable debugging at httplib level (requests->urllib3->http.client)
    # You will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
    # The only thing missing will be the response.body which is not logged.
    import http.client

    # TODO: nohow wrapped with logging, plain prints (heh heh), so formatting will not be consistent
    http.client.HTTPConnection.debuglevel = 1

    # for requests we can define logging properly
    requests_log = LoggerHelper(logtarget="requests.packages.urllib3").get_initialized_logger()
    requests_log.setLevel(lgr.getEffectiveLevel())
    requests_log.propagate = True

__docformat__ = 'restructuredtext'


def process_www_authenticate(v):
    if not v:
        return []
    # TODO: provide proper parsing/handling of this custom format and wider support:
    #   <type> realm=<realm>[, charset="UTF-8"]
    # More notes: https://github.com/datalad/datalad/issues/5846#issuecomment-890221053
    # The most complete solution is from 2018 on https://stackoverflow.com/a/52462292/1265472
    # relying on parsing it using pyparsing.
    supported_type = v.split(' ')[0].lower()
    our_type = {
        'basic': 'http_basic_auth',
        'digest': 'http_digest_auth',
        # TODO: bearer_token_anon ?
    }.get(supported_type)
    return [our_type] if our_type else []


def check_response_status(response, err_prefix="", session=None):
    """Check if response's status_code signals problem with authentication etc

    ATM succeeds only if response code was 200
    """
    if not err_prefix:
        err_prefix = "Access to %s has failed: " % response.url
    # 401 would be for digest authentication mechanism, or if we first ask which mechanisms are
    # supported.... must be linked into the logic if we decide to automagically detect which
    # mechanism or to give more sensible error message
    err_msg = err_prefix + "status code %d" % response.status_code
    if response.status_code in {404}:
        # It could have been that form_url is wrong, so let's just say that
        # TODO: actually may be that is where we could use tagid and actually determine the form submission url
        raise DownloadError(err_prefix + "not found")
    elif 400 <= response.status_code < 500:
        raise AccessDeniedError(
            err_msg,
            supported_types=process_www_authenticate(
                response.headers.get('WWW-Authenticate')),
            status=response.status_code)
    elif response.status_code in {200}:
        pass
    elif response.status_code in {301, 302, 307}:
        # TODO: apparently tests do not exercise this one yet
        if session is None:
            raise AccessFailedError(err_msg + " no session was provided",
                                    status=response.status_code)
        redirs = list(session.resolve_redirects(response, response.request))
        if len(redirs) > 1:
            lgr.warning("Multiple redirects aren't supported yet.  Taking first")
        elif len(redirs) == 0:
            raise AccessFailedError("No redirects were resolved",
                                    status=response.status_code)
        raise UnhandledRedirectError(err_msg, url=redirs[0].url,
                                     status=response.status_code)
    else:
        raise AccessFailedError(err_msg, status=response.status_code)


@auto_repr
class HTTPBaseAuthenticator(Authenticator):
    """Base class for html_form and http_auth authenticators
    """
    def __init__(self, url=None, failure_re=None, success_re=None,
                 session_cookies=None, **kwargs):
        """
        Parameters
        ----------
        url : str, optional
          URL where to find the form/login to authenticate.  If not provided, an original query url
          which will be provided to the __call__ of the authenticator will be used
        failure_re : str or list of str, optional
        success_re : str or list of str, optional
          Regular expressions to determine if login has failed or succeeded.
          TODO: we might condition when it gets ran
        session_cookies : str or list of str, optional
          Session cookies to store (besides auth response cookies)
        """
        super(HTTPBaseAuthenticator, self).__init__(**kwargs)
        self.url = url
        self.failure_re = ensure_list_from_str(failure_re)
        self.success_re = ensure_list_from_str(success_re)
        self.session_cookies = ensure_list_from_str(session_cookies)

    def authenticate(self, url, credential, session, update=False):
        # we should use specified URL for this authentication first
        lgr.info("http session: Authenticating into session for %s", url)
        post_url = self.url if self.url else url
        credential.set_context(auth_url=post_url)
        credentials = credential()

        # The whole thing relies on server first spitting out 401
        # and client getting again with 'Authentication:' header
        # So we need custom handling for those, while keeping track not
        # of cookies per se, but of 'Authentication:' header which is
        # to be used in subsequent GETs
        response = self._post_credential(credentials, post_url, session)
        if response is None:
            # authentication did not involve any interaction, nothing to
            # check at this point
            return

        # Handle responses if there was initial authentication exchange,
        # e.g. posting to a form and getting a cookie etc
        err_prefix = "Authentication to %s failed: " % post_url
        try:
            check_response_status(response, err_prefix, session=session)
        except DownloadError:
            # It might have happened that the return code was 'incorrect'
            # and we did get some feedback, which we could analyze to
            # figure out actual problem.  E.g. in case of nersc of crcns
            # it returns 404 (not found) with text in the html
            if response is not None and response.text:
                self.check_for_auth_failure(response.text, err_prefix)
            raise

        response_text = response.text
        self.check_for_auth_failure(response_text, err_prefix)

        if self.success_re:
            # the one which must be used to verify success
            # verify that we actually logged in
            for success_re in self.success_re:
                if not re.search(success_re, response_text):
                    raise AccessDeniedError(
                        err_prefix + " returned output did not match 'success' regular expression %s" % success_re
                    )

        cookies_dict = {}
        if response.cookies:
            cookies_dict = requests.utils.dict_from_cookiejar(response.cookies)
        if self.session_cookies:
            # any session cookies to store
            cookies_dict.update({k: session.cookies[k] for k in self.session_cookies})

        if cookies_dict:
            if (url in cookies_db) and update:
                cookies_db[url].update(cookies_dict)
            else:
                cookies_db[url] = cookies_dict
            # assign cookies for this session
            for c, v in cookies_dict.items():
                if c not in session.cookies or session.cookies[c] != v:
                    session.cookies[c] = v  # .update(cookies_dict)

        return response

    def _post_credential(self, credentials, post_url, session):
        raise NotImplementedError("Must be implemented in subclass")

    def check_for_auth_failure(self, content, err_prefix=""):
        if self.failure_re:
            content_is_bytes = isinstance(content, bytes)
            # verify that we actually logged in
            for failure_re in self.failure_re:
                if content_is_bytes:
                    # content could be not in utf-8. But I do not think that
                    # it is worth ATM messing around with guessing encoding
                    # of the content to figure out what to encode it into
                    # since typically returned "auth failed" should be in
                    # utf-8 or plain ascii
                    failure_re = ensure_bytes(failure_re)
                if re.search(failure_re, content):
                    raise AccessDeniedError(
                        err_prefix + "returned output which matches regular expression %s" % failure_re
                    )


@auto_repr
class HTMLFormAuthenticator(HTTPBaseAuthenticator):
    """Authenticate by opening a session via POSTing to HTML form
    """

    def __init__(self, fields, tagid=None, **kwargs):
        """

        Example specification in the .ini config file
        [provider:crcns]
        ...
        credential = crcns ; is not given to authenticator as is
        authentication_type = html_form
        # TODO: may be rename into post_url
        html_form_url = https://crcns.org/login_form
        # probably not needed actually since form_url
        # html_form_tagid = login_form
        html_form_fields = __ac_name={user}
                   __ac_password={password}
                   submit=Log in
                   form.submitted=1
                   js_enabled=0
                   cookies_enabled=
        html_form_failure_re = (Login failed|Please log in)
        html_form_success_re = You are now logged in

        Parameters
        ----------
        fields : str or dict
          String or a dictionary, which will be used (along with credential) information
          to feed into the form
        tagid : str, optional
          id of the HTML <form> in the document to use. If None, and page contains a single form,
          that one will be used.  If multiple forms -- error will be raise
        **kwargs : dict, optional
          Passed to super class HTTPBaseAuthenticator
        """
        super(HTMLFormAuthenticator, self).__init__(**kwargs)
        self.fields = ensure_dict_from_str(fields)
        self.tagid = tagid

    def _post_credential(self, credentials, post_url, session):
        post_fields = {
            k: v.format(**credentials)
            for k, v in self.fields.items()
        }

        response = session.post(post_url, data=post_fields)
        lgr.debug("Posted to %s fields %s, got response %s with headers %s",
                  post_url, list(post_fields.keys()), response,
                  list(response.headers.keys()))
        return response


@auto_repr
class HTTPRequestsAuthenticator(HTTPBaseAuthenticator):
    """Base class for various authenticators using requests pre-crafted ones


    Note, that current implementation assumes REQUESTS_FIELDS to be identical to
    the keys of a `Credential` object's FIELDS.
    """

    REQUESTS_AUTHENTICATOR = None
    REQUESTS_FIELDS = ('user', 'password')

    def __init__(self, **kwargs):
        # so we have __init__ solely for a custom docstring
        super(HTTPRequestsAuthenticator, self).__init__(**kwargs)

    def _post_credential(self, credentials, post_url, session):
        authenticator = self.REQUESTS_AUTHENTICATOR(
            *[credentials[f] for f in self.REQUESTS_FIELDS])
        session.auth = authenticator


@auto_repr
class HTTPBasicAuthAuthenticator(HTTPRequestsAuthenticator):
    """Authenticate via basic HTTP authentication

    Example specification in the .ini config file
    [provider:hcp-db]
    ...
    credential = hcp-db
    authentication_type = http_auth

    Parameters
    ----------
    **kwargs : dict, optional
      Passed to super class HTTPBaseAuthenticator
    """

    REQUESTS_AUTHENTICATOR = requests.auth.HTTPBasicAuth


@auto_repr
class HTTPAuthAuthenticator(HTTPRequestsAuthenticator):
    """Authenticate via Basic authentication to some other post url

    TODO:  actually this is some remnants which might later were RFed
    into the form authenticator since otherwise they make little sense
    """

    REQUESTS_AUTHENTICATOR = requests.auth.HTTPBasicAuth

    def _post_credential(self, credentials, post_url, session):
        authenticator = self.REQUESTS_AUTHENTICATOR(
            *[credentials[f] for f in self.REQUESTS_FIELDS])
        session.auth = authenticator
        response = session.post(post_url, data={},
                                auth=authenticator)
        auth_request = response.headers.get('www-authenticate')
        if response.status_code == 401 and auth_request:
            if auth_request.lower().split(' ', 1)[0] == 'basic':
                if response.url != post_url:
                    # was instructed to authenticate elsewhere
                    # TODO: do we need to loop may be??
                    response2 = session.get(response.url, auth=authenticator)
                    return response2
            else:
                lgr.warning(
                    f"{self} received response with www-authenticate={auth_request!r} "
                    "which is not Basic, and thus it cannot handle ATM.")
        return response


@auto_repr
class HTTPDigestAuthAuthenticator(HTTPRequestsAuthenticator):
    """Authenticate via HTTP digest authentication
    """

    REQUESTS_AUTHENTICATOR = requests.auth.HTTPDigestAuth


@auto_repr
class HTTPBearerTokenAuthenticator(HTTPRequestsAuthenticator):
    """Authenticate via HTTP 'Authorization: Bearer TOKEN' header

    E.g. as defined for OAuth2 in RFC 6750
    https://datatracker.ietf.org/doc/html/rfc6750
    """

    DEFAULT_CREDENTIAL_TYPE = 'token'
    AUTH_KEYWORD = 'Bearer'

    def __init__(self, **kwargs):
        # so we have __init__ solely for a custom docstring
        super(HTTPBearerTokenAuthenticator, self).__init__(**kwargs)

    def _post_credential(self, credentials, post_url, session):
        # we do not need to post anything, just inject token into the session
        session.headers['Authorization'] = f"{self.AUTH_KEYWORD} {credentials['token']}"


class HTTPTokenAuthenticator(HTTPBearerTokenAuthenticator):
    """Authenticate via HTTP 'Authorization: Token TOKEN' header

    It is pretty much the "Bearer TOKEN" method but which uses different keyword
    "Token".  It is e.g. the one provided by Django REST Framework.
    GitHub allows for both 'Bearer' and 'Token' keywords:
    https://docs.github.com/en/rest/authentication/authenticating-to-the-rest-api?apiVersion=2022-11-28
    """
    AUTH_KEYWORD = 'Token'


@auto_repr
class HTTPAnonBearerTokenAuthenticator(HTTPBearerTokenAuthenticator):
    """Retrieve token via 401 response and add Authorization: Bearer header.
    """

    allows_anonymous = True

    def authenticate(self, url, credential, session, update=False):
        if credential:
            lgr.warning(
                "Argument 'credential' specified, but it will be ignored: %s",
                credential)
        response = session.head(url)
        status = response.status_code
        if status == 200:
            lgr.debug("No authorization needed for %s", url)
            return
        if status != 401:
            raise DownloadError(
                "Expected 200 or 401 but got {} from {}"
                .format(status, url))

        lgr.debug("Requesting authorization token for %s", url)
        # TODO: it is not RFC 2068 Section 2 format, but a custom
        # <type> realm=<realm>[, charset="UTF-8"]
        # see TODO/harmonize with  process_www_authenticate
        auth_parts = parse_dict_header(response.headers["www-authenticate"])
        auth_url = ("{}?service={}&scope={}"
                    .format(auth_parts["Bearer realm"],
                            auth_parts["service"],
                            auth_parts["scope"]))
        auth_response = session.get(auth_url)
        try:
            auth_info = auth_response.json()
        except ValueError as e:
            raise DownloadError(
                "Failed to get information from {}"
                .format(auth_url)) from e
        session.headers['Authorization'] = "Bearer " + auth_info["token"]


@auto_repr
class HTTPDownloaderSession(DownloaderSession):
    def __init__(self, size=None, filename=None,  url=None, headers=None,
                 response=None, chunk_size=1024 ** 2):
        super(HTTPDownloaderSession, self).__init__(
            size=size, filename=filename, url=url, headers=headers,
        )
        self.chunk_size = chunk_size
        self.response = response

    def download(self, f=None, pbar=None, size=None):
        response = self.response
        # content_gzipped = 'gzip' in response.headers.get('content-encoding', '').split(',')
        # if content_gzipped:
        #     raise NotImplemented("We do not support (yet) gzipped content")
        #     # see https://rationalpie.wordpress.com/2010/06/02/python-streaming-gzip-decompression/
        #     # for ways to implement in python 2 and 3.2's gzip is working better with streams

        total = 0
        return_content = f is None
        if f is None:
            # no file to download to
            # TODO: actually strange since it should have been decoded then...
            f = io.BytesIO()

        # must use .raw to be able avoiding decoding/decompression while downloading
        # to a file
        chunk_size_ = min(self.chunk_size, size) if size is not None else self.chunk_size

        # XXX With requests_ftp BytesIO is provided as response.raw for ftp urls,
        # which has no .stream, so let's do ducktyping and provide our custom stream
        # via BufferedReader for such cases, while maintaining the rest of code
        # intact.  TODO: figure it all out, since doesn't scale for any sizeable download
        # This code is tested by tests/test_http.py:test_download_ftp BUT
        # it causes 503 on travis,  but not always so we allow to skip that test
        # in such cases. That causes fluctuating coverage
        if not hasattr(response.raw, 'stream'):  # pragma: no cover
            def _stream():
                buf = io.BufferedReader(response.raw)
                v = True
                while v:
                    v = buf.read(chunk_size_)
                    yield v

            stream = _stream()
        else:
            # XXX TODO -- it must be just a dirty workaround
            # As we discovered with downloads from NITRC all headers come with
            # Content-Encoding: gzip which leads  requests to decode them.  But the point
            # is that ftp links (yoh doesn't think) are gzip compressed for the transfer
            decode_content = not response.url.startswith('ftp://')
            stream = response.raw.stream(chunk_size_, decode_content=decode_content)

        for chunk in stream:
            if chunk:  # filter out keep-alive new chunks
                chunk_len = len(chunk)
                if size is not None and total + chunk_len > size:
                    # trim the download to match target size
                    chunk = chunk[:size - total]
                    chunk_len = len(chunk)
                total += chunk_len
                f.write(chunk)
                try:
                    # TODO: pbar is not robust ATM against > 100% performance ;)
                    if pbar:
                        pbar.update(total)
                except Exception as e:
                    ce = CapturedException(e)
                    lgr.warning("Failed to update progressbar: %s", ce)
                # TEMP
                # see https://github.com/niltonvolpato/python-progressbar/pull/44
                ui.out.flush()
                if size is not None and total >= size:  # pragma: no cover
                    break  # we have done as much as we were asked

        if return_content:
            out = f.getvalue()
            return out


@auto_repr
class HTTPDownloader(BaseDownloader):
    """A stateful downloader to maintain a session to the website
    """

    @borrowkwargs(BaseDownloader)
    def __init__(self, headers=None, **kwargs):
        """

        Parameters
        ----------
        headers: dict, optional
          Header fields to be provided to the session. Unless User-Agent provided, a custom
          one, available in `DEFAULT_USER_AGENT` constant of this module will be used.
        """
        super(HTTPDownloader, self).__init__(**kwargs)
        self._session = None
        headers = headers.copy() if headers else {}
        if 'user-agent' not in map(str.lower, headers):
            headers['User-Agent'] = DEFAULT_USER_AGENT
        self._headers = headers

    def _establish_session(self, url, allow_old=True):
        """

        Parameters
        ----------
        allow_old: bool, optional
          If a Downloader allows for persistent sessions by some means -- flag
          instructs whether to use previous session, or establish a new one

        Returns
        -------
        bool
          To state if old instance of a session/authentication was used
        """
        if allow_old:
            if self._session:
                lgr.debug("http session: Reusing previous")
                return True  # we used old
            elif url in cookies_db:
                cookie_dict = cookies_db[url]
                lgr.debug("http session: Creating new with old cookies %s", list(cookie_dict.keys()))
                self._session = requests.Session()
                # not sure what happens if cookie is expired (need check to that or exception will prolly get thrown)

                # TODO dict_to_cookiejar doesn't preserve all fields when reversed
                self._session.cookies = requests.utils.cookiejar_from_dict(cookie_dict)
                # TODO cookie could be expired w/ something like (but docs say it should be expired automatically):
                # http://docs.python-requests.org/en/latest/api/#requests.cookies.RequestsCookieJar.clear_expired_cookies
                # self._session.cookies.clear_expired_cookies()
                return True

        lgr.debug("http session: Creating brand new session")
        self._session = requests.Session()
        self._session.headers.update(self._headers)
        if self.authenticator:
            self.authenticator.authenticate(url, self.credential, self._session)

        return False

    def get_downloader_session(self, url,
                               allow_redirects=True,
                               use_redirected_url=True,
                               headers=None):
        # TODO: possibly make chunk size adaptive
        # TODO: make it not this ugly -- but at the moment we are testing end-file size
        # while can't know for sure if content was gunziped and whether it all went ok.
        # So safer option -- just request to not have it gzipped
        if headers is None:
            headers = {}
        if 'Accept-Encoding' not in headers:
            headers['Accept-Encoding'] = ''

        # TODO: our tests ATM aren't ready for retries, thus altogether disabled for now
        nretries = 1
        for retry in range(1, nretries+1):
            try:
                response = self._session.get(
                    url, stream=True, allow_redirects=allow_redirects,
                    headers=headers)
            #except (MaxRetryError, NewConnectionError) as exc:
            except Exception as exc:
                ce = CapturedException(exc)
                # happen to run into those with urls pointing to Amazon,
                # so let's rest and try again
                if retry >= nretries:
                    #import epdb; epdb.serve()
                    if not _FTP_SUPPORT and url.startswith("ftp://"):
                        msg_ftp = "For ftp:// support, install requests_ftp. "
                    else:
                        msg_ftp = ""

                    raise AccessFailedError(
                        "Failed to establish a new session %d times. %s"
                        % (nretries, msg_ftp)) from exc
                lgr.warning(
                    "Caught exception %s. Will retry %d out of %d times",
                    ce, retry + 1, nretries)
                sleep(2**retry)

        check_response_status(response, session=self._session)
        headers = response.headers
        lgr.debug("Establishing session for url %s, response headers: %s",
                  url, headers)
        target_size = int(headers.get('Content-Length', '0').strip()) or None
        if use_redirected_url and response.url and response.url != url:
            lgr.debug("URL %s was redirected to %s and thus the later will be used"
                      % (url, response.url))
            url = response.url
        # Consult about filename.  Since we already have headers,
        # should not result in an additional request
        url_filename = get_url_filename(url, headers=headers)

        headers['Url-Filename'] = url_filename
        return HTTPDownloaderSession(
            size=target_size,
            url=response.url,
            filename=url_filename,
            headers=headers,
            response=response
        )

    @classmethod
    def get_status_from_headers(cls, headers):
        """Given HTTP headers, return 'status' record to assess later if link content was changed
        """
        # used for quick checks for HTTP or S3?
        # TODO:  So we will base all statuses on this set? e.g. for Last-Modified if to be
        # mapping from field to its type converter
        HTTP_HEADERS_TO_STATUS = {
            'Content-Length': int,
            'Content-Disposition': str,
            'Last-Modified': rfc2822_to_epoch,
            'Url-Filename': str,
        }
        # Allow for webserver to return them in other casing
        HTTP_HEADERS_TO_STATUS_lower = {s.lower(): (s, t) for s, t in HTTP_HEADERS_TO_STATUS.items()}
        status = {}
        if headers:
            for header_key in headers:
                try:
                    k, t = HTTP_HEADERS_TO_STATUS_lower[header_key.lower()]
                except KeyError:
                    continue
                status[k] = t(headers[header_key])

        # convert to FileStatus
        return FileStatus(
            size=status.get('Content-Length'),
            mtime=status.get('Last-Modified'),
            filename=get_response_disposition_filename(
                status.get('Content-Disposition')) or status.get('Url-Filename')
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Provide access to Amazon S3 objects.
"""

import logging
import re
from logging import getLogger
from threading import Lock
from urllib.parse import unquote as urlunquote
from urllib.parse import urlsplit

import boto3
import botocore

from ..dochelpers import borrowkwargs
from ..support.exceptions import (
    AccessDeniedError,
    AccessPermissionExpiredError,
    TargetFileAbsent,
)
from ..support.network import get_url_straight_filename
from ..support.status import FileStatus
from ..utils import (
    auto_repr,
    ensure_dict_from_str,
)
from .base import (
    Authenticator,
    BaseDownloader,
    DownloaderSession,
)

lgr = getLogger('datalad.s3')
boto_lgr = logging.getLogger('boto3')
# not in effect at all, probably those are setup later
#boto_lgr.handlers = lgr.handlers  # Use our handlers

import warnings

__docformat__ = 'restructuredtext'


@auto_repr
class S3Authenticator(Authenticator):
    """Authenticator for S3 AWS
    """
    allows_anonymous = True
    DEFAULT_CREDENTIAL_TYPE = 'aws-s3'

    def __init__(self, *args, host=None, region=None, **kwargs):
        """

        Parameters
        ----------
        host: str, optional
          A URL to the endpoint or just a host name (then prepended with https://).
          Define either `host` or `region`, not both.
        region: str, optional
          will be passed to s3 client as region_name.
          Default region could also be defined in the `~/.aws/config`
          file or by setting the AWS_DEFAULT_REGION environment variable.
        """
        super(S3Authenticator, self).__init__(*args, **kwargs)
        self.client = None
        self.region = region
        if host:
            if region:
                raise ValueError("Define either 'host' or 'region', not both.")
            if not (host.startswith('http://') or host.startswith('https://')):
                host = 'https://' + host
        self.host = host

    def authenticate(self, bucket_name, credential, cache=True):
        """Authenticates to the specified bucket using provided credentials

        Sets up a boto3 S3 client. Uses supplied DataLad credentials
        or connects anonymously - in that order of preference.

        No test is done to verify bucket access, to avoid hitting the case
        where a bucket cannot be listed but its objects can be accessed.

        Returns
        -------
        s3client: botocore.client.S3
        """

        if not boto3:
            raise RuntimeError("%s requires boto3 module which is N/A" % self)

        # Shut up boto if we do not care to listen ;)
        boto_lgr.setLevel(
            logging.CRITICAL
            if lgr.getEffectiveLevel() > 1
            else logging.DEBUG
        )

        # use boto3 standard retry mode, not legacy
        conf_options = {"retries": {"mode": "standard"}}

        # per http://stackoverflow.com/a/19089045/1265472 & updated for boto3
        if bucket_name.lower() != bucket_name or "." in bucket_name:
            conf_options["s3"] = {"addressing_style": "path"}

        conf = botocore.config.Config(**conf_options)

        # credential might contain 'session' token as well
        # which could be provided   as   security_token=<token>.,
        # see http://stackoverflow.com/questions/7673840/is-there-a-way-to-create-a-s3-connection-with-a-sessions-token

        if credential is not None:
            credentials = credential()
            conn_kind = "with authentication"
            s3client = boto3.client(
                "s3",
                region_name=self.region,
                endpoint_url=self.host,
                aws_access_key_id=credentials["key_id"],
                aws_secret_access_key=credentials["secret_id"],
                aws_session_token=credentials.get("session"),
                config=conf,
            )
        else:
            # Boto3 has mechanisms to read credentials from the environment
            # or configuration files.
            # Signed requests for publicly accessible objects may fail, so
            # for now we will assume that anonymous access is preferred.
            session = boto3.Session()
            conn_kind = "anonymously"
            conf = conf.merge(
                botocore.config.Config(signature_version=botocore.UNSIGNED)
            )
            s3client = session.client("s3", region_name=self.region, config=conf)

        lgr.info(
            "S3 session: Connecting to the bucket %s %s", bucket_name, conn_kind
        )

        self.client = s3client
        return s3client


@auto_repr
class S3DownloaderSession(DownloaderSession):
    def __init__(self, size=None, filename=None, url=None, headers=None,
                 client=None,
                 bucket: str=None, key: str=None, version_kwargs: dict=None):
        super(S3DownloaderSession, self).__init__(
            size=size, filename=filename, headers=headers, url=url
        )
        self.client = client
        self.bucket = bucket
        self.key = key
        self.version_kwargs = version_kwargs
        self.pbar_callback_lock = Lock()

    def download(self, f=None, pbar=None, size=None):
        # S3 specific (the rest is common with e.g. http)
        def pbar_callback(downloaded):
            with self.pbar_callback_lock:
                if pbar:
                    try:
                        pbar.update(downloaded, increment=True)
                    except:  # MIH: what does it do? MemoryError?
                        pass  # do not let pbar spoil our fun

        if f:
            if size is None:
                # TODO: May be we could use If-Modified-Since
                # see http://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectGET.html
                # note (MSz): in boto3, see client.head_object(IfModifiedSince)
                self.client.download_fileobj(
                    Fileobj=f,
                    Callback=pbar_callback,
                    Bucket=self.bucket,
                    Key=self.key,
                    ExtraArgs=self.version_kwargs,
                )
                return
            # The problem is that there is no Range support in download_fileobj
            # https://github.com/boto/boto3/issues/1215 ,
            # so we have to use get_object, and will just save it into a file.
            # It will not work for large transfers, but hopefully we do not have them
            # for file downloads

        # return the contents of the file as bytes
        # ATM no progress indication!
        kwargs = dict(Bucket=self.bucket, Key=self.key, **self.version_kwargs)
        if size is not None:
            # This will work only for relatively small transfers but
            # in general we do not expect requests with large-ish size.
            kwargs['Range'] = f'bytes=0-{size - 1}'
        s3_response = self.client.get_object(
            **kwargs,
            # There is no callback for get_object!
            # Callback=pbar_callback,
        )
        content = s3_response.get('Body').read()
        if f:
            f.write(content)
        else:
            return content


@auto_repr
class S3Downloader(BaseDownloader):
    """Downloader from AWS S3 buckets
    """

    _DEFAULT_AUTHENTICATOR = S3Authenticator

    @borrowkwargs(BaseDownloader)
    def __init__(self, **kwargs):
        super(S3Downloader, self).__init__(**kwargs)
        self._client = None
        self._bucket_name = None

    @property
    def client(self):
        return self._client

    def reset(self):
        self._client = None
        self._bucket_name = None

    @classmethod
    def _parse_url(cls, url, bucket_only=False):
        """Parses s3:// url and returns bucket name, prefix, additional query elements
         as a dict (such as VersionId)"""
        rec = urlsplit(url)
        if bucket_only:
            return rec.netloc
        assert(rec.scheme == 's3')
        # We are often working with urlencoded URLs so we could safely interact
        # with git-annex via its text based protocol etc.  So, if URL looks like
        # it was urlencoded the filepath, we should revert back to an original key
        # name.  Since we did not demarcate whether it was urlencoded, we will do
        # magical check, which would fail if someone had % followed by two digits
        filepath = rec.path.lstrip('/')
        if re.search('%[0-9a-fA-F]{2}', filepath):
            lgr.debug("URL unquoting S3 URL filepath %s", filepath)
            filepath = urlunquote(filepath)
        # TODO: needs replacement to ensure_ since it doesn't
        # deal with non key=value
        return rec.netloc, filepath, ensure_dict_from_str(rec.query, sep='&') or {}

    def _establish_session(self, url, allow_old=True):
        """

        Parameters
        ----------
        allow_old: bool, optional
          If a Downloader allows for persistent sessions by some means -- flag
          instructs whether to use previous session, or establish a new one

        Returns
        -------
        bool
          To state if old instance of a session/authentication was used
        """
        bucket_name = self._parse_url(url, bucket_only=True)
        if allow_old and self.client is not None:
            if self._bucket_name == bucket_name:
                try:
                    self._check_credential()
                    lgr.debug(
                        "S3 session: Reusing previous connection to bucket %s",
                        bucket_name
                    )
                    return True  # we used old
                except AccessPermissionExpiredError:
                    lgr.debug("S3 session: credential expired")
            else:
                lgr.warning("No support yet for multiple buckets per S3Downloader")

        lgr.debug("S3 session: Reconnecting to the bucket")
        self._client = self.authenticator.authenticate(
            bucket_name, self.credential)
        self._bucket_name = bucket_name
        return False

    def _check_credential(self):
        """Quick check of the credential if known on either it has not expired

        Raises
        ------
        AccessPermissionExpiredError
          if credential is found to be expired
        """
        if self.credential and self.credential.is_expired:
            raise AccessPermissionExpiredError(
                "Credential %s has expired" % self.credential)

    def get_downloader_session(self, url, **kwargs):
        """Create a DownloaderSession for a given URL

        This function sets up a DownloaderSession object and reads the
        information necessary (e.g. size, headers) by issuing a
        head_object query.

        Returns
        -------
        S3DownloaderSession
        """
        bucket_name, url_filepath, params = self._parse_url(url)
        if params:
            newkeys = set(params.keys()) - {'versionId'}
            if newkeys:
                raise NotImplementedError("Did not implement support for %s" % newkeys)
            # see: boto3.s3.transfer.S3Transfer.ALLOWED_DOWNLOAD_ARGS
            version_kwargs = {"VersionId": params.get("versionId")}
        else:
            version_kwargs = {}

        assert(self._bucket_name == bucket_name)  # must be the same

        self._check_credential()

        # this is where the *real* access check (for the object) will happen;
        # may raise botocore.exceptions.ClientError for 404 not found
        # or 403 forbidden
        try:
            object_meta = self.client.head_object(
                Bucket=bucket_name,
                Key=url_filepath,
                **version_kwargs,
            )
        except botocore.exceptions.ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "403":
                raise AccessDeniedError(e) from e
            elif error_code == "404":
                raise TargetFileAbsent(url + " does not exist") from e
            elif error_code == '400' and version_kwargs:
                try:
                    self.client.head_object(
                        Bucket=bucket_name,
                        Key=url_filepath,
                    )
                    raise TargetFileAbsent(url + " has unknown version specified") from e
                except botocore.exceptions.ClientError as e2:
                    raise e
            else:
                raise

        target_size = object_meta.get('ContentLength')  # S3 specific
        headers = self.get_obj_headers(
            object_meta, other={'Content-Disposition': url_filepath})

        # Consult about filename
        url_filename = get_url_straight_filename(url)

        # head_object also reports VersionId if found, but we don't
        # take it from there, only from the URL; original comment by
        # yoh below
        #
        # It is a good idea in general to avoid race between moment of retrieving
        # the key information and actual download.
        # But depending on permissions, we might be unable (like in the case with NDA)
        # to download a guaranteed version of the key.
        # So we will just download the latest version (if still there)
        # if no versionId was specified in URL
        # Alternative would be to make this a generator and generate sessions
        # but also remember if the first download succeeded so we do not try
        # again to get versioned one first.
        # TODO: ask NDA to allow download of specific versionId?

        return S3DownloaderSession(
            size=target_size,
            filename=url_filename,
            url=url,
            headers=headers,
            client=self.client,
            bucket=bucket_name,
            key=url_filepath,
            version_kwargs=version_kwargs,
        )

    @classmethod
    def get_obj_headers(cls, obj_meta, other=None):
        """Get a headers dict from head_object output

        Picks, renames, and converts certain keys from the boto3
        head_object response. Some (like Content-Disposition) may need
        to be added via kwargs.

        Note: the head_object output also includes the relevant
        information under ['ResponseMetadata']['HTTPHeaders']
        (possibly using different data type), but this function only
        uses top-level keys.
        """
        headers = {"Content-Length": obj_meta.get("ContentLength")}
        if obj_meta.get("LastModified"):
            headers["Last-Modified"] = int(
                obj_meta.get("LastModified").timestamp())
        if other:
            headers.update(other)
        return headers

    @classmethod
    def get_status_from_headers(cls, headers):
        # TODO: duplicated with http functionality
        # convert to FileStatus
        return FileStatus(
            size=headers.get('Content-Length'),
            mtime=headers.get('Last-Modified'),
            filename=headers.get('Content-Disposition')
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Adapters and decorators for vcr
"""

import logging
import sys
from contextlib import contextmanager
from functools import wraps
from os.path import isabs

from datalad.support.exceptions import CapturedException
from datalad.utils import Path

lgr = logging.getLogger("datalad.support.vcr")


def _caller_cassette_path(name):
    """Look for ``vcr_cassettes/<name>.yaml`` next to the test that called us.

    `use_cassette` is typically applied as a decorator at module import time,
    so the first frame outside this module is the test module that wants
    the cassette.  When used as a context manager from inside a test
    function the same logic still resolves to that test's module file.
    Returns the absolute path on hit, or None if no such cassette exists.
    """
    frame = sys._getframe(1)
    here = __file__
    while frame is not None:
        caller = frame.f_code.co_filename
        if caller and caller != here:
            candidate = Path(caller).parent / "vcr_cassettes" / f"{name}.yaml"
            if candidate.is_file():
                return str(candidate)
            return None
        frame = frame.f_back
    return None


def _get_cassette_path(path):
    """Return a path to the cassette within our 'storage'.

    An absolute path is used verbatim.  A bare name is resolved against:

    1. ``<caller-module-dir>/vcr_cassettes/<name>.yaml`` (versioned, shipped
       alongside the test that requested it), if such a file exists.
    2. otherwise ``fixtures/vcr_cassettes/<name>.yaml`` relative to the
       current working directory -- the legacy gitignored scratch location
       where VCR's ``record_mode='once'`` writes a fresh cassette on the
       first run.  Useful for local recording before promoting the cassette
       into the versioned per-test directory.
    """
    if isabs(path):
        return path
    versioned = _caller_cassette_path(path)
    if versioned is not None:
        return versioned
    return "fixtures/vcr_cassettes/%s.yaml" % path

try:
    # TEMP: Just to overcome problem with testing on jessie with older requests
    # https://github.com/kevin1024/vcrpy/issues/215
    import requests as _
    import vcr.patch as _vcrp
    try:
        from requests.packages.urllib3.connectionpool import \
            HTTPConnection as _a
        from requests.packages.urllib3.connectionpool import \
            VerifiedHTTPSConnection as _b
    except ImportError:
        def returnnothing(*args, **kwargs):
            return()
        _vcrp.CassettePatcherBuilder._requests = returnnothing

    from vcr import VCR as _VCR
    from vcr import use_cassette as _use_cassette

    def use_cassette(path, return_body=None, skip_if_no_vcr=False, **kwargs):
        """Adapter so we could create/use custom use_cassette with custom parameters

        Parameters
        ----------
        path : str
          If not absolute path, treated as a name for a cassette under fixtures/vcr_cassettes/
        skip_if_no_vcr : bool
          Rather than running without VCR it would throw unittest.SkipTest
          exception.  Of effect only if vcr import fails (so not in this
          implementation but the one below)
        """
        path = _get_cassette_path(path)
        lgr.debug("Using cassette %s", path)
        if return_body is not None:
            my_vcr = _VCR(
                before_record_response=lambda r: dict(r, body={'string': return_body.encode()}))
            return my_vcr.use_cassette(path, **kwargs)  # with a custom response
        else:
            return _use_cassette(path, **kwargs)  # just a straight one

    # shush vcr
    vcr_lgr = logging.getLogger('vcr')
    if lgr.getEffectiveLevel() > logging.DEBUG:
        vcr_lgr.setLevel(logging.WARN)
except Exception as exc:
    if not isinstance(exc, ImportError):
        # something else went hairy (e.g. vcr failed to import boto due to some syntax error)
        lgr.warning("Failed to import vcr, no cassettes will be available: %s",
                    CapturedException(exc))
    # If there is no vcr.py -- provide a do nothing decorator for use_cassette

    def use_cassette(path, return_body=None, skip_if_no_vcr=False, **kwargs):
        if skip_if_no_vcr:
            def skip_decorator(t):
                @wraps(t)
                def wrapper(*args, **kwargs):
                    from unittest import SkipTest
                    raise SkipTest("No vcr")
                return wrapper
            return skip_decorator
        else:
            def do_nothing_decorator(t):
                @wraps(t)
                def wrapper(*args, **kwargs):
                    lgr.debug("Not using vcr cassette")
                    return t(*args, **kwargs)
                return wrapper
            return do_nothing_decorator


@contextmanager
def externals_use_cassette(name):
    """Helper to pass instruction via env variables to use specified cassette

    For instance whenever we are testing custom special remotes invoked by the annex
    but want to minimize their network traffic by using vcr.py
    """
    from unittest.mock import patch
    cassette_path = str(Path(_get_cassette_path(name)).resolve())  # realpath OK
    with patch.dict('os.environ', {'DATALAD_TESTS_USECASSETTE': cassette_path}):
        yield



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Data providers - bind downloaders and credentials together
"""

import re
from configparser import ConfigParser as SafeConfigParserWithIncludes
from glob import glob
from logging import getLogger
from os.path import (
    abspath,
    dirname,
)
from os.path import join as pathjoin
from urllib.parse import urlparse

from datalad.support import path
from datalad.support.external_versions import external_versions
from datalad.support.network import RI
from datalad.utils import (
    Path,
    auto_repr,
    ensure_list_from_str,
    get_dataset_root,
)

from ..interface.common_cfg import dirs
from .base import (
    NoneAuthenticator,
    NotImplementedAuthenticator,
)
from .http import (
    HTMLFormAuthenticator,
    HTTPAnonBearerTokenAuthenticator,
    HTTPAuthAuthenticator,
    HTTPBasicAuthAuthenticator,
    HTTPBearerTokenAuthenticator,
    HTTPDigestAuthAuthenticator,
    HTTPDownloader,
    HTTPTokenAuthenticator,
)
from .s3 import (
    S3Authenticator,
    S3Downloader,
)
from .shub import SHubDownloader

lgr = getLogger('datalad.downloaders.providers')

# dict to bind authentication_type's to authenticator classes
# parameters will be fetched from config file itself
AUTHENTICATION_TYPES = {
    'html_form': HTMLFormAuthenticator,
    'http_auth': HTTPAuthAuthenticator,
    'http_basic_auth': HTTPBasicAuthAuthenticator,
    'http_digest_auth': HTTPDigestAuthAuthenticator,
    'http_token': HTTPTokenAuthenticator,
    'bearer_token': HTTPBearerTokenAuthenticator,
    'bearer_token_anon': HTTPAnonBearerTokenAuthenticator,
    'aws-s3': S3Authenticator,  # TODO: check if having '-' is kosher
    'nda-s3': S3Authenticator,
    'loris-token': HTTPBearerTokenAuthenticator,
    'xnat': NotImplementedAuthenticator,
    'none': NoneAuthenticator,
}

from datalad.downloaders import CREDENTIAL_TYPES


@auto_repr
class Provider(object):
    """Class to bring together url_res, credential, and authenticator
    """
    # TODO: we might need a lazy loading of the submodules which would provide
    # specific downloaders while importing needed Python modules "on demand"
    DOWNLOADERS = {
        'http': {'class': HTTPDownloader, 'externals': {'requests'}},
        'https': {'class': HTTPDownloader, 'externals': {'requests'}},
        'shub': {'class': SHubDownloader, 'externals': {'requests'}},
        'ftp': {'class': HTTPDownloader, 'externals': {'requests', 'boto3'}},
        's3': {'class': S3Downloader, 'externals': {'boto3'}}
        # ... TODO
    }

    def __init__(self, name, url_res, credential=None, authenticator=None,
                 downloader=None):
        """
        Parameters
        ----------
        name: str
        url_res: list of str
           Regular expressions
        credential: Credential, optional
        authenticator: Authenticator, optional
        downloader: Downloader, optional

        """
        self.name = name
        self.url_res = ensure_list_from_str(url_res)
        self.credential = credential
        self.authenticator = authenticator
        self._downloader = downloader

    @property
    def downloader(self):
        return self._downloader

    @staticmethod
    def get_scheme_from_url(url):
        """Given a URL return scheme to decide which downloader class to use
        """
        url_split = urlparse(url)
        return url_split.scheme  # , url_split.netloc)

    @classmethod
    def _get_downloader_class(cls, url):
        key = cls.get_scheme_from_url(url)
        if key in cls.DOWNLOADERS:
            entry = cls.DOWNLOADERS[key]
            klass = entry['class']
            for ext in entry.get('externals', set()):
                if external_versions[ext] is None:
                    raise RuntimeError(
                        "For using %s downloader, you need '%s' dependency "
                        "which seems to be missing" % (klass, ext)
                    )
            return klass
        else:
            raise ValueError("Do not know how to handle url %s for scheme %s. Known: %s"
                             % (url, key, cls.DOWNLOADERS.keys()))

    def get_downloader(self, url, **kwargs):
        """Assigns proper downloader given the URL

        If one is known -- verifies its appropriateness for the given url.
        ATM we do not support multiple types of downloaders per single provider
        """
        if self._downloader is None:
            # we need to create a new one
            Downloader = self._get_downloader_class(url)
            # we might need to provide it with credentials and authenticator
            # Let's do via kwargs so we could accommodate cases when downloader does not necessarily
            # cares about those... duck typing or what it is in action
            kwargs = kwargs.copy()
            if self.credential:
                kwargs['credential'] = self.credential
            if self.authenticator:
                kwargs['authenticator'] = self.authenticator
            self._downloader = Downloader(**kwargs)
        return self._downloader


class Providers(object):
    """

    So we could provide handling for URLs with corresponding credentials
    and specific (reusable) downloader.  Internally it contains
    Providers and interfaces them based on a given URL.  Each provider
    in turn takes care about associated with it Downloader.
    """

    _DEFAULT_PROVIDERS = None
    _DS_ROOT = None
    _CONFIG_TEMPLATE = """\
# Provider configuration file created to initially access
# {url}

[provider:{name}]
url_re = {url_re}
authentication_type = {authentication_type}
# Note that you might need to specify additional fields specific to the
# authenticator.  Fow now "look into the docs/source" of {authenticator_class}
# {authentication_type}_
credential = {credential_name}

[credential:{credential_name}]
# If known, specify URL or email to how/where to request credentials
# url = ???
type = {credential_type}
"""

    def __init__(self, providers=None):
        """
        """
        self._providers = providers or []
        # a set of providers to handle connections without authentication.
        # Will be setup one per each protocol schema
        self._default_providers = {}

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            "" if not self._providers else repr(self._providers)
        )

    def __len__(self):
        return len(self._providers)

    def __getitem__(self, index):
        return self._providers[index]

    def __iter__(self):
        return self._providers.__iter__()

    @classmethod
    def _get_providers_dirs(cls, dsroot=None):
        """Return an ordered dict with directories to look for provider config files

        Is implemented as a function to ease mock testing depending on dirs.
        values
        """
        paths = dict()
        paths['dist'] = pathjoin(dirname(abspath(__file__)), 'configs')
        if dsroot is not None:
            paths['ds'] = pathjoin(dsroot, '.datalad', 'providers')
        paths['site'] = pathjoin(dirs.site_config_dir, "providers") \
            if dirs.site_config_dir else None
        paths['user'] = pathjoin(dirs.user_config_dir, "providers") \
            if dirs.user_config_dir else None
        return paths

    @classmethod
    def _get_configs(cls, dir, files='*.cfg'):
        return glob(pathjoin(dir, files)) if dir is not None else []

    @classmethod
    def from_config_files(cls, files=None, reload=False):
        """Loads information about related/possible websites requiring authentication from:

        - datalad/downloaders/configs/*.cfg files provided by the codebase
        - current dataset .datalad/providers/
        - User's home directory directory (ie ~/.config/datalad/providers/*.cfg)
        - system-wide datalad installation/config (ie /etc/datalad/providers/*.cfg)

        For sample configs files see datalad/downloaders/configs/providers.cfg

        If files is None, loading is cached between calls.  Specify reload=True to force
        reloading of files from the filesystem.  The class method reset_default_providers
        can also be called to reset the cached providers.
        """
        # lazy part
        # Note, that "" is effectively the same as ".". If CWD is in a dataset's
        # root already, `get_dataset_root` will just return its input. While
        # this is technically correct, we want to use it here in order to
        # figure, whether the dsroot is the same as when we called last time.
        # However, "", ".", etc. would stay the same, whenever we switch into
        # another dataset, causing `from_config_files` to "think" nothing
        # changed. Therefore use abspath here.
        dsroot_rel = get_dataset_root("")
        dsroot = abspath(dsroot_rel) if dsroot_rel is not None else None
        if files is None and cls._DEFAULT_PROVIDERS and not reload and dsroot==cls._DS_ROOT:
            return cls._DEFAULT_PROVIDERS

        config = SafeConfigParserWithIncludes()
        files_orig = files
        if files is None:
            cls._DS_ROOT = dsroot
            files = []
            for p in cls._get_providers_dirs(dsroot).values():
                files.extend(cls._get_configs(p))
        config.read(files)

        # We need first to load Providers and credentials
        # Order matters, because we need to ensure that when
        # there's a conflict between configuration files declared
        # at different precedence levels (ie. dataset vs system)
        # the appropriate precedence config wins.
        providers = dict()
        credentials = {}

        for section in config.sections():
            if ':' in section:
                type_, name = section.split(':', 1)
                assert type_ in {'provider', 'credential'}, "we know only providers and credentials, got type %s" % type_
                items = {
                    o: config.get(section, o) for o in config.options(section)
                }
                # side-effect -- items get popped
                locals().get(type_ + "s")[name] = getattr(
                    cls, '_process_' + type_)(name, items)
                if len(items):
                    raise ValueError("Unprocessed fields left for %s: %s" % (name, str(items)))
            else:
                lgr.warning("Do not know how to treat section %s here" % section)

        # link credentials into providers
        lgr.debug("Assigning credentials into %d providers", len(providers))
        for provider in providers.values():
            if provider.credential:
                if provider.credential not in credentials:
                    raise ValueError("Unknown credential %s. Known are: %s"
                                     % (provider.credential, ", ".join(credentials.keys())))
                provider.credential = credentials[provider.credential]
                # TODO: Is this the right place to pass dataset to credential?
                provider.credential.set_context(dataset=cls._DS_ROOT)

        providers = Providers(list(providers.values()))

        if files_orig is None:
            # Store providers for lazy access
            cls._DEFAULT_PROVIDERS = providers

        return providers

    @classmethod
    def reset_default_providers(cls):
        """Resets to None memoized by from_config_files providers
        """
        cls._DEFAULT_PROVIDERS = None

    @classmethod
    def _process_provider(cls, name, items):
        """Process a dictionary specifying the provider and output the Provider instance
        """
        assert 'authentication_type' in items, "Must have authentication_type specified"

        auth_type = items.pop('authentication_type')
        if auth_type not in AUTHENTICATION_TYPES:
            raise ValueError("Unknown authentication_type=%s. Known are: %s"
                             % (auth_type, ', '.join(AUTHENTICATION_TYPES)))

        if auth_type != 'none':
            authenticator = AUTHENTICATION_TYPES[auth_type](
                # Extract all the fields as keyword arguments
                **{k[len(auth_type) + 1:]: items.pop(k)
                   for k in list(items.keys())
                   if k.startswith(auth_type + "_")}
            )
        else:
            authenticator = None

        # bringing url_re to "standard" format of a list and populating _providers_ordered
        url_res = ensure_list_from_str(items.pop('url_re', []))
        assert url_res, "current implementation relies on having url_re defined"

        credential = items.pop('credential', None)

        # credential instance will be assigned later after all of them are loaded
        return Provider(name=name, url_res=url_res, authenticator=authenticator,
                        credential=credential)

    @classmethod
    def _process_credential(cls, name, items):
        assert 'type' in items, "Credential must specify type.  Missing in %s" % name
        cred_type = items.pop('type')
        if cred_type not in CREDENTIAL_TYPES:
            raise ValueError("I do not know type %s credential. Known: %s"
                             % (cred_type, CREDENTIAL_TYPES.keys()))
        return CREDENTIAL_TYPES[cred_type](name=name, url=items.pop('url', None))

    def reload(self):
        new_providers = self.from_config_files(reload=True)
        self._providers = new_providers._providers
        self._default_providers = new_providers._default_providers

    def get_provider(self, url, only_nondefault=False, return_all=False):
        """Given a URL returns matching provider
        """

        # Range backwards to ensure that more locally defined
        # configuration wins in conflicts between url_re
        matching_providers = []
        for provider in self._providers[::-1]:
            for url_re in provider.url_res:
                try:
                    if re.match(url_re, url):
                        lgr.debug("Returning provider %s for url %s", provider, url)
                        matching_providers.append(provider)
                except re.error:
                    lgr.warning(
                        "Invalid regex %s in provider %s"
                        % (url_re, provider.name)
                    )

        if matching_providers:
            if return_all:
                return matching_providers
            if len(matching_providers) > 1:
                lgr.warning(
                    "Multiple providers matched for %s, using the first one"
                    % url)
            return matching_providers[0]

        if only_nondefault:
            return None

        # None matched -- so we should get a default one per each of used
        # protocols
        scheme = Provider.get_scheme_from_url(url)
        if scheme not in self._default_providers:
            lgr.debug("Initializing default provider for %s", scheme)
            self._default_providers[scheme] = Provider(name="", url_res=["%s://.*" % scheme])
        provider = self._default_providers[scheme]
        lgr.debug("No dedicated provider, returning default one for %s: %s",
                  scheme, provider)
        return provider

    def _store_new(self, url=None, authentication_type=None,
                   authenticator_class=None, url_re=None, name=None,
                   credential_name=None, credential_type=None, level='user'):
        """Stores a provider and credential config and reloads afterwards.

        Note
        ----
        non-interactive version of `enter_new`.
        For now non-public, pending further refactoring

        Parameters
        ----------
        level: str
          Where to store the config. Choices: 'user' (default), 'ds', 'site'

        Returns
        -------
        Provider
          The stored `Provider` as reported by reload
        """

        # We don't ask user for confirmation, so for this non-interactive
        # routine require everything to be explicitly specified.
        if any(not a for a in [url, authentication_type, authenticator_class,
                               url_re, name, credential_name, credential_type]):
            raise ValueError("All arguments must be specified")

        if level not in ['user', 'ds', 'site']:
            raise ValueError("'level' must be one of 'user', 'ds', 'site'")

        providers_dir = Path(self._get_providers_dirs()[level])
        if not providers_dir.exists():
            providers_dir.mkdir(parents=True, exist_ok=True)
        filepath = providers_dir / f"{name}.cfg"
        cfg = self._CONFIG_TEMPLATE.format(**locals())
        filepath.write_bytes(cfg.encode('utf-8'))
        self.reload()
        return self.get_provider(url)

    def enter_new(self, url=None, auth_types=[], url_re=None, name=None,
                  credential_name=None, credential_type=None):
        # TODO: level/location!
        """Create new provider and credential config

        If interactive, this will ask the user to enter the details (or confirm
        default choices). A dedicated config file is written at
        <user_config_dir>/providers/<name>.cfg

        Parameters:
        -----------
        url: str or RI
          URL this config is created for
        auth_types: list
          List of authentication types to choose from. First entry becomes
          default. See datalad.downloaders.providers.AUTHENTICATION_TYPES
        url_re: str
          regular expression; Once created, this config will be used for any
          matching URL; defaults to `url`
        name: str
          name for the provider; needs to be unique per user
        credential_name: str
          name for the credential; defaults to the provider's name
        credential_type: str
          credential type to use (key for datalad.downloaders.CREDENTIAL_TYPES)
        """

        from datalad.ui import ui
        if url and not name:
            ri = RI(url)
            for f in ('hostname', 'name'):
                try:
                    # might need sanitarization
                    name = str(getattr(ri, f))
                except AttributeError:
                    pass
        known_providers_by_name = {p.name: p for p in self._providers}
        providers_user_dir = self._get_providers_dirs()['user']
        while True:
            name = ui.question(
                title="New provider name",
                text="Unique name to identify 'provider' for %s" % url,
                default=name
            )
            filename = pathjoin(providers_user_dir, '%s.cfg' % name)
            if name in known_providers_by_name:
                if ui.yesno(
                    title="Known provider %s" % name,
                    text="Provider with name %s already known. Do you want to "
                         "use it for this session?"
                         % name,
                    default=True
                ):
                    return known_providers_by_name[name]
            elif path.lexists(filename):
                ui.error(
                    "File %s already exists, choose another name" % filename)
            else:
                break

        if not credential_name:
            credential_name = name
        if not url_re:
            url_re = re.escape(url) if url else None
        while True:
            url_re = ui.question(
                title="New provider regular expression",
                text="A (Python) regular expression to specify for which URLs "
                     "this provider should be used",
                default=url_re
            )
            if not re.match(url_re, url):
                ui.error("Provided regular expression doesn't match original "
                         "url.  Please re-enter")
            # TODO: url_re of another provider might match it as well
            #  I am not sure if we have any kind of "priority" setting ATM
            #  to differentiate or to to try multiple types :-/
            else:
                break

        authentication_type = None
        if auth_types:
            auth_types = [
                t for t in auth_types if t in AUTHENTICATION_TYPES
            ]
            if auth_types:
                authentication_type = auth_types[0]

        # Setup credential
        authentication_type = ui.question(
            title="Authentication type",
            text="What authentication type to use",
            default=authentication_type,
            choices=sorted(AUTHENTICATION_TYPES)
        )
        authenticator_class = AUTHENTICATION_TYPES[authentication_type]

        # TODO: need to figure out what fields that authenticator might
        #       need to have setup and ask for them here!

        credential_type = ui.question(
            title="Credential",
            text="What type of credential should be used?",
            choices=sorted(CREDENTIAL_TYPES),
            default=credential_type or getattr(authenticator_class,
                                               'DEFAULT_CREDENTIAL_TYPE')
        )

        cfg = self._CONFIG_TEMPLATE.format(**locals())
        if ui.yesno(
            title="Save provider configuration file",
            text="Following configuration will be written to %s:\n%s"
                % (filename, cfg),
            default='yes'
        ):
            # Just create a configuration file and reload the thing
            return self._store_new(url=url,
                                   authentication_type=authentication_type,
                                   authenticator_class=authenticator_class,
                                   url_re=url_re,
                                   name=name,
                                   credential_name=credential_name,
                                   credential_type=credential_type,
                                   level='user'
                                   )
        else:
            return None


    # TODO: avoid duplication somehow ;)
    # Sugarings to get easier access to downloaders
    def download(self, url, *args, **kwargs):
        return self.get_provider(url).get_downloader(url).download(url, *args, **kwargs)

    def fetch(self, url, *args, **kwargs):
        return self.get_provider(url).get_downloader(url).fetch(url, *args, **kwargs)

    def get_status(self, url, *args, **kwargs):
        return self.get_provider(url).get_downloader(url).get_status(url, *args, **kwargs)

    def needs_authentication(self, url):
        provider = self.get_provider(url, only_nondefault=True)
        if provider is None:
            return None
        return provider.authenticator is not None


    # # TODO: UNUSED?
    # def get_credentials(self, url, new=False):
    #     """Ask user to enter credentials for a provider matching url
    #     """
    #     # find a match among _items
    #     provider = self.get_provider(url)
    #     if new or not provider:
    #         rec = self._get_new_record_ui(url)
    #         rec['url_re'] = "TODO"  # present to user and ask to edit
    #         name = urlparse(url).netloc
    #         self._items[name] = rec
    #         if ui.yesno("Do you want to store credentials for %s" % name):
    #             self.store_credentials()
    #     else:
    #         return self._items[name]
    #
    # def store_credentials(self, name):
    #     # TODO: store  self._items[name]  in appropriate (user) creds
    #     # for later reuse
    #     raise NotImplementedError()
    #
    # def _get_new_record_ui(self, url):
    #     # TODO: should be a dialog with the fields appropriate for this particular
    #     # type of credentials
    #     ui.message("To access %s we would need credentials." % url)
    #     if url in self.providers:
    #         ui.message("If you don't yet have credentials, please visit %s"
    #                    % self.providers.get(url, 'credentials_url'))
    #     return { 'user': ui.question("Username:"),
    #              'password': ui.password() }



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tests for vcr adapter"""

from pathlib import Path

from ...tests.utils_pytest import (
    SkipTest,
    eq_,
)
from ..vcr_ import (
    _get_cassette_path,
    use_cassette,
)


def test_get_cassette_path_absolute(tmp_path):
    # Absolute paths are returned unchanged
    abs_path = str(tmp_path / "foo.yaml")
    eq_(_get_cassette_path(abs_path), abs_path)


def test_get_cassette_path_caller_relative(tmp_path, monkeypatch):
    # When a versioned cassette exists next to the calling module under
    # `vcr_cassettes/<name>.yaml`, that path is returned.
    cassette_dir = Path(__file__).parent / "vcr_cassettes"
    cassette_dir.mkdir(exist_ok=True)
    cassette = cassette_dir / "test_get_cassette_path_caller_relative.yaml"
    cassette.write_text("interactions: []\nversion: 1\n")
    try:
        eq_(_get_cassette_path("test_get_cassette_path_caller_relative"),
            str(cassette))
    finally:
        cassette.unlink()
        # only remove the dir if we created it for this test
        try:
            cassette_dir.rmdir()
        except OSError:
            pass


def test_get_cassette_path_fallback(tmp_path, monkeypatch):
    # No caller-relative cassette -> legacy CWD-relative scratch path.
    # Use a unique-enough name so that no test shipping a cassette by
    # that name would shadow this assertion.
    name = "datalad_vcr_test_definitely_no_such_cassette_xyz"
    eq_(_get_cassette_path(name),
        "fixtures/vcr_cassettes/%s.yaml" % name)


def test_use_cassette_if_no_vcr():
    # just test that our do nothing decorator does the right thing if vcr is not present
    skip = False
    try:
        import vcr
        skip = True
    except ImportError:
        pass
    except:
        # if anything else goes wrong with importing vcr, we still should be able to
        # run use_cassette
        pass
    if skip:
        raise SkipTest("vcr is present, can't test behavior with vcr presence ATM")

    @use_cassette("some_path")
    def checker(x):
        return x + 1

    eq_(checker(1), 2)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import sys
from typing import (
    Concatenate,
    Literal,
    ParamSpec,
    Protocol,
    TypedDict,
    TypeVar,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

__all__ = ["Literal", "ParamSpec", "T", "K", "V", "P"]

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")
P = ParamSpec("P")



# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test WitlessRunner
"""

from __future__ import annotations

import logging
import os
import signal
import sys
import unittest.mock
from threading import (
    Lock,
    Thread,
)
from time import (
    sleep,
    time,
)
from typing import Any

import pytest

from datalad.tests.utils_pytest import (
    FILESYSTEM_SUPPORTS_UTF8,
    OBSCURE_FILENAME,
    SkipTest,
    assert_cwd_unchanged,
    assert_in,
    assert_raises,
    eq_,
    integration,
    ok_,
    ok_file_has_content,
    skip_if_on_windows,
    swallow_logs,
    with_tempfile,
)
from datalad.utils import (
    CMD_MAX_ARG,
    Path,
    on_windows,
)

from .. import (
    CommandError,
    KillOutput,
    Protocol,
    Runner,
    StdOutCapture,
    StdOutErrCapture,
)
from .utils import py2cmd

result_counter = 0


@assert_cwd_unchanged
@with_tempfile
def test_runner(tempfile: str = "") -> None:
    runner = Runner()
    content = 'Testing real run' if on_windows or not FILESYSTEM_SUPPORTS_UTF8 else 'Testing äöü東 real run'
    cmd = 'echo %s > %s' % (content, tempfile)
    res = runner.run(cmd)
    assert isinstance(res, dict)
    # no capture of any kind, by default
    ok_(not res['stdout'])
    ok_(not res['stderr'])
    ok_file_has_content(tempfile, content, strip=True)
    os.unlink(tempfile)


def test_runner_stderr_capture() -> None:
    runner = Runner()
    test_msg = "stderr-Message"
    res = runner.run(py2cmd(
        'import sys; print(%r, file=sys.stderr)' % test_msg),
        protocol=StdOutErrCapture,
    )
    assert isinstance(res, dict)
    eq_(res['stderr'].rstrip(), test_msg)
    ok_(not res['stdout'])


def test_runner_stdout_capture() -> None:
    runner = Runner()
    test_msg = "stdout-Message"
    res = runner.run(py2cmd(
        'import sys; print(%r, file=sys.stdout)' % test_msg),
        protocol=StdOutErrCapture,
    )
    assert isinstance(res, dict)
    eq_(res['stdout'].rstrip(), test_msg)
    ok_(not res['stderr'])


def test_runner_failure() -> None:
    runner = Runner()
    with assert_raises(CommandError) as cme:
        runner.run(
            py2cmd('import sys; sys.exit(53)')
        )
    eq_(53, cme.value.code)

    # but we bubble up FileNotFoundError if executable does not exist at all
    with assert_raises(FileNotFoundError) as cme:
        runner.run(['dne1l2k3j4'])  # be damned the one who makes such a command


@with_tempfile(mkdir=True)
def test_runner_fix_PWD(path: str = "") -> None:
    env = os.environ.copy()
    env['PWD'] = orig_cwd = os.getcwd()
    runner = Runner(cwd=path, env=env)
    res = runner.run(
        py2cmd('import os; print(os.environ["PWD"])'),
        protocol=StdOutCapture,
    )
    assert isinstance(res, dict)
    eq_(res['stdout'].strip(), path)  # was fixed up to point to point to cwd's path
    eq_(env['PWD'], orig_cwd)  # no side-effect


@with_tempfile(mkdir=True)
def test_runner_cwd_encoding(path: str = "") -> None:
    env = os.environ.copy()
    # Add PWD to env so that runner will temporarily adjust it to point to cwd.
    env['PWD'] = os.getcwd()
    cwd = Path(path) / OBSCURE_FILENAME
    cwd.mkdir()
    # Running doesn't fail if cwd or env has unicode value.
    Runner(cwd=cwd, env=env).run(
        py2cmd(
            'from pathlib import Path; (Path.cwd() / "foo").write_text("t")'))
    (cwd / 'foo').exists()


@with_tempfile(mkdir=True)
def test_runner_stdin(path: str = "") -> None:
    runner = Runner()
    fakestdin = Path(path) / 'io'
    # go for difficult content
    fakestdin.write_text(OBSCURE_FILENAME)

    res = runner.run(
        py2cmd('import fileinput; print(fileinput.input().readline())'),
        stdin=fakestdin.open(),
        protocol=StdOutCapture,
    )
    assert isinstance(res, dict)
    assert_in(OBSCURE_FILENAME, res['stdout'])

    # we can do the same without a tempfile, too
    res = runner.run(
        py2cmd('import fileinput; print(fileinput.input().readline())'),
        stdin=OBSCURE_FILENAME.encode('utf-8'),
        protocol=StdOutCapture,
    )
    assert isinstance(res, dict)
    assert_in(OBSCURE_FILENAME, res['stdout'])


@pytest.mark.fail_slow(3)
def test_runner_stdin_no_capture() -> None:
    # Ensure that stdin writing alone progresses
    runner = Runner()
    runner.run(
        py2cmd('import sys; print(sys.stdin.read()[-10:])'),
        stdin=('ABCDEFGHIJKLMNOPQRSTUVWXYZ-' * 2 + '\n').encode('utf-8'),
        protocol=None
    )


@pytest.mark.fail_slow(3)
def test_runner_no_stdin_no_capture() -> None:
    # Ensure a runner without stdin data and output capture progresses
    runner = Runner()
    runner.run(
        (["cmd.exe", "/c"] if on_windows else []) + ["echo", "a", "b", "c"],
        stdin=None,
        protocol=None
    )


@pytest.mark.fail_slow(3)
def test_runner_empty_stdin() -> None:
    # Ensure a runner without stdin data and output capture progresses
    runner = Runner()
    runner.run(
        py2cmd('import sys; print(sys.stdin.read())'),
        stdin=b"",
        protocol=None
    )


def test_runner_parametrized_protocol() -> None:
    runner = Runner()

    # protocol returns a given value whatever it receives
    class ProtocolInt(StdOutCapture):
        def __init__(self, value: bytes) -> None:
            self.value = value
            super().__init__()

        def pipe_data_received(self, fd: int, data: bytes) -> None:
            super().pipe_data_received(fd, self.value)

    res = runner.run(
        py2cmd('print(1, end="")'),
        protocol=ProtocolInt,
        # value passed to protocol constructor
        value=b'5',
    )
    assert isinstance(res, dict)
    eq_(res['stdout'], '5')


@integration  # ~3 sec
@with_tempfile(mkdir=True)
@with_tempfile()
def test_asyncio_loop_noninterference1(path1: str = "", path2: str = "") -> None:
    # minimalistic use case provided by Dorota
    import datalad.api as dl
    src = dl.create(path1)  # type: ignore[attr-defined]
    reproducer = src.pathobj/ "reproducer.py"
    reproducer.write_text(f"""\
import asyncio
import sys
# Python 3.14+ no longer implicitly creates event loops in get_event_loop()
if sys.version_info >= (3, 14):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
else:
    asyncio.get_event_loop()
import datalad.api as datalad
ds = datalad.clone(path=r'{path2}', source=r"{path1}")
loop = asyncio.get_event_loop()
assert loop
# simulate outside process closing the loop
loop.close()
# and us still doing ok
ds.status()
""")
    Runner().run([sys.executable, str(reproducer)])  # if Error -- the test failed


@with_tempfile
def test_asyncio_forked(temp_: str = "") -> None:
    # temp will be used to communicate from child either it succeeded or not
    temp = Path(temp_)
    runner = Runner()
    try:
        pid = os.fork()
    except BaseException as exc:
        # .fork availability is "Unix", and there are cases where it is "not supported"
        # so we will just skip if no forking is possible
        raise SkipTest(f"Cannot fork: {exc}")
    # if does not fail (in original or in a fork) -- we are good
    try:
        runner.run([sys.executable, '--version'], protocol=StdOutCapture)
        if pid == 0:
            temp.write_text("I rule")
    except:
        if pid == 0:
            temp.write_text("I suck")
    if pid != 0:
       # parent: look after the child
       t0 = time()
       try:
           while not temp.exists() or temp.stat().st_size < 6:
               if time() - t0 > 5:
                   raise AssertionError("Child process did not create a file we expected!")
       finally:
           # kill the child
           os.kill(pid, signal.SIGTERM)
       # see if it was a good one
       eq_(temp.read_text(), "I rule")
    else:
       # sleep enough so parent just kills me the kid before I continue doing bad deeds
       sleep(10)


def test_done_deprecation() -> None:
    with unittest.mock.patch("datalad.cmd.warnings.warn") as warn_mock:
        _ = Protocol("done")
        warn_mock.assert_called_once()

    with unittest.mock.patch("datalad.cmd.warnings.warn") as warn_mock:
        _ = Protocol()
        warn_mock.assert_not_called()


def test_faulty_poll_detection() -> None:
    popen_mock = unittest.mock.MagicMock(**{"pid": 666, "poll.return_value": None})
    protocol = Protocol()
    protocol.process = popen_mock
    assert_raises(CommandError, protocol._prepare_result)


def test_kill_output() -> None:
    runner = Runner()
    res = runner.run(
        py2cmd('import sys; sys.stdout.write("aaaa\\n"); sys.stderr.write("bbbb\\n")'),
        protocol=KillOutput)
    assert isinstance(res, dict)
    eq_(res['stdout'], '')
    eq_(res['stderr'], '')


@skip_if_on_windows  # no "hint" on windows since no ulimit command there
def test_too_long() -> None:
    with swallow_logs(new_level=logging.ERROR) as cml:
        with assert_raises(OSError):  # we still raise an exception if we exceed too much
            Runner().run(
                [sys.executable, '-c', 'import sys; print(len(sys.argv))'] + [str(i) for i in range(CMD_MAX_ARG)],
                protocol=StdOutCapture
            )
        cml.assert_logged('.*use.*ulimit.*')


def test_path_to_str_conversion() -> None:
    # Regression test to ensure that Path-objects are converted into strings
    # before they are put into the environment variable `$PWD`
    runner = Runner()
    test_path = Path("a/b/c")
    adjusted_env = runner._get_adjusted_env(
        cwd=test_path,
        env=dict(some_key="value")
    )
    assert adjusted_env is not None
    assert str(test_path) == adjusted_env['PWD']


def test_env_copying() -> None:
    # Regression test to ensure environments are only copied
    # if `copy=True` is given to `Runner._get_adjusted_env.`
    # Test also for path adjustments, if not-`None` `pwd`-value
    # is given to `Runner._get_adjusted_env`.
    runner = Runner()
    for original_env in (None, dict(some_key='value')):
        for cwd in (None, Path('a/b/c')):
            for do_copy in (True, False):
                adjusted_env = runner._get_adjusted_env(
                    cwd=cwd,
                    env=original_env,
                    copy=do_copy
                )
                if original_env is None:
                    assert adjusted_env is None
                else:
                    assert adjusted_env is not None
                    if do_copy is True:
                        assert adjusted_env is not original_env
                    else:
                        assert adjusted_env is original_env
                    if cwd is None:
                        assert 'PWD' not in adjusted_env
                    else:
                        assert 'PWD' in adjusted_env


@with_tempfile(mkdir=True)
def test_environment(temp_dir_path: str = "") -> None:
    # Ensure that the subprocess sees a string in `$PWD`, even if a Path-object
    # is provided to `cwd`.
    cmd = py2cmd("import os; print(os.environ['PWD'])")
    cwd = Path(temp_dir_path)
    env = dict(SYSTEMROOT=os.environ.get('SYSTEMROOT', ''))
    runner = Runner()
    results = runner.run(cmd=cmd, protocol=StdOutCapture, cwd=cwd, env=env)
    assert isinstance(results, dict)
    output = results['stdout'].splitlines()[0]
    assert output == temp_dir_path

    runner = Runner(cwd=cwd, env=env)
    results = runner.run(cmd=cmd, protocol=StdOutCapture)
    assert isinstance(results, dict)
    output = results['stdout'].splitlines()[0]
    assert output == temp_dir_path


def test_argument_priority() -> None:
    class X:
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            self.args = args
            self.kwargs = kwargs

        def run(self) -> dict:
            return dict(
                code=0,
                args=self.args,
                kwargs=self.kwargs,
            )

    test_path_1 = "a/b/c"
    test_env_1 = dict(source="constructor")
    test_path_2 = "d/e/f"
    test_env_2 = dict(source="run-method")

    with unittest.mock.patch('datalad.runner.runner.ThreadedRunner') as tr_mock:

        tr_mock.side_effect = X
        runner = Runner(cwd=test_path_1, env=test_env_1)

        result = runner.run("first-command")
        assert isinstance(result, dict)
        assert result['kwargs']['cwd'] == test_path_1
        assert result['kwargs']['env'] == {
            **test_env_1,
            'PWD': test_path_1
        }

        result = runner.run("second-command", cwd=test_path_2, env=test_env_2)
        assert isinstance(result, dict)
        assert result['kwargs']['cwd'] == test_path_2
        assert result['kwargs']['env'] == {
            **test_env_2,
            'PWD': test_path_2
        }


def test_concurrent_execution() -> None:
    runner = Runner()
    caller_threads = []

    result_list: list[str] = []
    result_list_lock = Lock()

    def target(count: int, r_list: list[str], r_list_lock: Lock) -> None:
        result = runner.run(
            py2cmd(
                "import time;"
                "import sys;"
                "time.sleep(1);"
                "print('end', sys.argv[1])",
                str(count)
            ),
            protocol=StdOutCapture,
        )
        assert isinstance(result, dict)
        output = result["stdout"].strip()
        assert output == f"end {str(count)}"
        with r_list_lock:
            r_list.append(output)

    for c in range(100):
        caller_thread = Thread(
            target=target,
            kwargs=dict(
                count=c,
                r_list=result_list,
                r_list_lock=result_list_lock,
            ))
        caller_thread.start()
        caller_threads.append(caller_thread)

    while caller_threads:
        t = caller_threads.pop()
        t.join()

    assert len(result_list) == 100
